@@ -32,6 +32,8 @@
 #include        "gvaluecollector.h"
 #include	"gvaluetypes.h"
 #include	"gboxed.h"
+#include	"gobject.h"
+#include	"genums.h"
 
 #include	"gobjectalias.h"
 
@@ -226,9 +228,6 @@ static GBSearchConfig g_class_closure_bconfig = {
 static GHashTable    *g_handler_list_bsa_ht = NULL;
 static Emission      *g_recursive_emissions = NULL;
 static Emission      *g_restart_emissions = NULL;
-#ifndef DISABLE_MEM_POOLS
-static GTrashStack   *g_handler_ts = NULL;
-#endif
 static gulong         g_handler_sequential_number = 1;
 G_LOCK_DEFINE_STATIC (g_signal_mutex);
 #define	SIGNAL_LOCK()		G_LOCK (g_signal_mutex)
@@ -390,11 +389,6 @@ handler_match_prepend (HandlerMatch *list,
 {
   HandlerMatch *node;
   
-  /* yeah, we could use our own memchunk here, introducing yet more
-   * rarely used cached nodes and extra allocation overhead.
-   * instead, we use GList* nodes, since they are exactly the size
-   * we need and are already cached. g_signal_init() asserts this.
-   */
   node = g_slice_new (HandlerMatch);
   node->handler = handler;
   node->next = list;
@@ -717,9 +711,6 @@ g_signal_init (void)
   SIGNAL_LOCK ();
   if (!g_n_signal_nodes)
     {
-      /* handler_id_node_prepend() requires this */
-      g_assert (sizeof (GList) == sizeof (HandlerMatch));
-      
       /* setup handler list binary searchable array hash table (in german, that'd be one word ;) */
       g_handler_list_bsa_ht = g_hash_table_new (g_direct_hash, NULL);
       g_signal_key_bsa = g_bsearch_array_create (&g_signal_key_bconfig);
@@ -2616,8 +2607,6 @@ g_signal_accumulator_true_handled (GSignalInvocationHint *ihint,
 }
 
 /* --- compile standard marshallers --- */
-#include	"gobject.h"
-#include	"genums.h"
 #include        "gmarshal.c"
 
 #define __G_SIGNAL_C__
